
/*
 *  This file was generated by the SOM Compiler.
 *  Generated using:
 *     SOM incremental update: 2.24
 */


//***************************************************
// Lucide GBM plug-in
//
// Supported is:
// - loading all formats supported by GBM
// - scaled rendering
//
// Requires: GBM.DLL version 1.35 or higher (with multipage support)
//
// Author : Heiko Nitzsche
// Date   : 30-Jun-2006
// Version: 1.01 beta
//
// History: 27-Jun-2006: Initial public beta
//          30-Jun-2006: Public beta 2
//                       Fixes for the most urgent issues
//                       - file extension list too long for file dialog (only keep the first extension)
//                       - Add error message reporting to load_file())
//
//***************************************************

/* ***** BEGIN LICENSE BLOCK ***** */

/* This file is placed into the Public Domain. */

/* ***** END LICENSE BLOCK  ***** */

#ifndef SOM_Module_lugbm_Source
#define SOM_Module_lugbm_Source
#endif
#define LuGbmDocument_Class_Source

#include "lugbm.xih"

#define INCL_DOSMODULEMGR
#define INCL_DOSPROCESS
#include <os2.h>

#include <stdio.h>
#include <string.h>

#include "gbm.h"
#include "gbmscale.h"
#include "lugbmdyn.h"

// ---------------------------------------------------------
// ---------------------------------------------------------

#define PLUGIN_VERSION   "1.0 beta 2"

// ---------------------------------------------------------
// ---------------------------------------------------------

static char * supportedExtensions = NULL;

class GbmDocumentPage
{
    public:
      byte     * data;
      GBM        gbm;
      GBMRGB     gbmrgb[0x100];

      GbmDocumentPage()
        : data(NULL)
      {
          memset(&gbm  , 0, sizeof(GBM));
          memset(gbmrgb, 0, sizeof(gbmrgb));
      }

      ~GbmDocumentPage()
      {
          if (data != NULL)
          {
              free(data);
              data = NULL;
          }
      }
};

class GbmDocument
{
    public:
      GbmDocumentPage * pages;
      int               numPages;

      GbmDocument()
        : pages(NULL),
          numPages(0)
      {}

      ~GbmDocument()
      {
          deletePages();
      }

      void deletePages()
      {
          delete [] pages;
          pages    = NULL;
          numPages = 0;
      }
};

// ---------------------------------------------------------
// ---------------------------------------------------------

static BOOLEAN expand_to_24bit(const GBM *gbm, const GBMRGB *gbmrgb, const byte *data, byte *data24)
{
    const int stride     = ((gbm->w * gbm->bpp + 31)/32) * 4;
    const int new_stride = ((gbm->w * 3 + 3) & ~3);
          int y;

    if ( gbm->bpp == 24 )
    {
        return TRUE;
    }

    for ( y = 0; y < gbm->h; y++ )
    {
        const byte *src  = data + y * stride;
              byte *dest = data24 + y * new_stride;

        int x;

        switch ( gbm->bpp )
        {
            case 1:
            {
                byte c;

                for ( x = 0; x < gbm->w; x++ )
                {
                    if ( (x & 7) == 0 )
                        c = *src++;
                    else
                        c <<= 1;

                    *dest++ = gbmrgb[c >> 7].b;
                    *dest++ = gbmrgb[c >> 7].g;
                    *dest++ = gbmrgb[c >> 7].r;
                }
            }
            break;

            case 4:
            {
                for ( x = 0; x + 1 < gbm->w; x += 2 )
                {
                    const byte c = *src++;

                    *dest++ = gbmrgb[c >> 4].b;
                    *dest++ = gbmrgb[c >> 4].g;
                    *dest++ = gbmrgb[c >> 4].r;
                    *dest++ = gbmrgb[c & 15].b;
                    *dest++ = gbmrgb[c & 15].g;
                    *dest++ = gbmrgb[c & 15].r;
                    }

                if ( x < gbm->w )
                {
                    const byte c = *src;

                    *dest++ = gbmrgb[c >> 4].b;
                    *dest++ = gbmrgb[c >> 4].g;
                    *dest++ = gbmrgb[c >> 4].r;
                }
            }
            break;

            case 8:
            {
                for ( x = 0; x < gbm->w; x++ )
                {
                    const byte c = *src++;

                    *dest++ = gbmrgb[c].b;
                    *dest++ = gbmrgb[c].g;
                    *dest++ = gbmrgb[c].r;
                }
            }
            break;

            default:
               return FALSE;
        }
    }
    return TRUE;
}

// ---------------------------------------------------------

// Load a single bitmap page

static GBM_ERR load_gbm_page(char * filename, int fd, int ft, GBM * pGbm, GBMRGB * pGbmRgb, char * options,
                             byte ** data)
{
    GBM_ERR rc;

    rc = Gbm_read_header(filename, fd, ft, pGbm, options);
    if (rc != GBM_ERR_OK)
    {
        return rc;
    }

    rc = Gbm_read_palette(fd, ft, pGbm, pGbmRgb);
    if (rc != GBM_ERR_OK)
    {
        return rc;
    }

    // allocate the GBM loader buffer if bitmap is not 24bpp
    int    gbmBytes = 0;
    byte * gbmData  = NULL;

    if (pGbm->bpp != 24)
    {
        gbmBytes = ( ((pGbm->w * pGbm->bpp + 31)/32) * 4 ) * pGbm->h;
        gbmData  = (byte *) malloc(gbmBytes);
        if (gbmData == NULL)
        {
            return rc;
        }
    }

    const int gbmBytes24 = ( ((pGbm->w * 24 + 31)/32) * 4 ) * pGbm->h;
    *data = (byte *) malloc(gbmBytes24);
    if (*data == NULL)
    {
        return GBM_ERR_MEM;
    }

    if (pGbm->bpp != 24)
    {
        rc = Gbm_read_data(fd, ft, pGbm, gbmData);
        if (rc != GBM_ERR_OK)
        {
            free(gbmData);
            free(*data);
            gbmData = NULL;
            *data   = NULL;

            return rc;
        }

        // convert all color depths to 24 bits per pixel
        if (! expand_to_24bit(pGbm, pGbmRgb, gbmData, *data))
        {
            free(gbmData);
            free(*data);
            gbmData = NULL;
            *data   = NULL;

            return GBM_ERR_NOT_SUPP;
        }
        pGbm->bpp = 24;
        free(gbmData);
    }
    else
    {
        rc = Gbm_read_data(fd, ft, pGbm, *data);
        if (rc != GBM_ERR_OK)
        {
            free(*data);
            *data = NULL;

            return rc;
        }
    }

    return GBM_ERR_OK;
}

// ---------------------------------------------------------
// ---------------------------------------------------------

unsigned _System LibMain( unsigned hmod, unsigned termination )
{
    if ( termination )
    {
        /* DLL is detaching from process */
        if (supportedExtensions != NULL)
        {
            free(supportedExtensions);
            supportedExtensions = NULL;
        }
    }
    else
    {
        /* DLL is attaching to process */

        /* get all supported extensions from GBM */
        Gbm_init();

        int n_ft = 0;
        Gbm_query_n_filetypes(&n_ft);

        // we assume here that none of the import filters has more than 10 extensions with 3 chars
        int extensions_length = 10 * 3 * n_ft + 1;

        supportedExtensions = (char *) malloc(extensions_length);
        if (supportedExtensions == NULL)
        {
            Gbm_deinit();
            return 0;
        }
        strcpy(supportedExtensions, "");

        for (int ft = 0; ft < n_ft; ft++)
        {
            GBMFT gbmft;

            Gbm_query_filetype(ft, &gbmft);

            // get all extensions
            char * extTok = strtok(gbmft.extensions, " ");

#if 1 // Limit the number of reported extension to only the first one (most often used one)
      // until the file dialog extension list issue with the limited filename entry field
      // length is solved.

            if (extTok != NULL)
            {
                strcat(supportedExtensions, extTok);
                if (ft + 1 < n_ft)
                {
                    strcat(supportedExtensions, ";");
                }
            }
#else
            while (extTok != NULL)
            {
                if (strlen(supportedExtensions) + strlen(extTok) + 1 > extensions_length)
                {
                    // resize buffer
                    extensions_length += 10;
                    supportedExtensions = (char *) realloc(supportedExtensions, extensions_length);
                    if (supportedExtensions == NULL)
                    {
                        Gbm_deinit();
                        return 0;
                    }
                }
                strcat(supportedExtensions, extTok);
                extTok = strtok(NULL, " ");
                if (ft + 1 < n_ft)
                {
                    strcat(supportedExtensions, ";");
                }
                else
                {
                    if (extTok != NULL)
                    {
                        strcat(supportedExtensions, ";");
                    }
                }
            }
#endif
        }

        /* We might want to skip some formats like IAX that have no */
        /* header and thus usually need user input data which we don't */
        /* have here. */

        Gbm_deinit();
    }

    return( 1 );
}

// ---------------------------------------------------------

extern "C" LuDocument * _System createObject()
{
    return new LuGbmDocument();
}

// ---------------------------------------------------------

extern "C" char * _System getSupportedExtensions()
{
    return supportedExtensions;
}

// ---------------------------------------------------------

extern "C" char * _System getDescription()
{
    return "GBM plugin ("PLUGIN_VERSION"), based on Generalised Bitmap Module (GBM.DLL)";
}

// ---------------------------------------------------------
// ---------------------------------------------------------

SOM_Scope boolean  SOMLINK loadFile(LuGbmDocument *somSelf ,
                                    Environment   *ev      , string  filename,
                                    string         password, string* error)
{
    LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
    GbmDocument       *d       = (GbmDocument *)somThis->data;

    /* We read only the first page so far. */

    /* read the bitmap file */
    int     ft = -1;
    int     fd = -1;
    int     numPages = 1;
    GBM_ERR rc;

    Gbm_init();

    rc = Gbm_guess_filetype(filename, &ft);
    if (rc != GBM_ERR_OK)
    {
        const char * errorMessage       = Gbm_err(rc);
        const int    errorMessageLength = strlen(errorMessage);

        const char * message       = "Can't guess bitmap file format: %s";
        const int    messageLength = strlen(message);

        *error = (char *)SOMMalloc(messageLength + errorMessageLength + 1);
        sprintf(*error, message, errorMessage);

        Gbm_deinit();
        return FALSE;
    }

    if ( (fd = Gbm_io_open(filename, GBM_O_RDONLY)) == -1 )
    {
        const char * message       = "Can't open file.";
        const int    messageLength = strlen(message);

        *error = (char *)SOMMalloc(messageLength + 1);
        strcpy(*error, message);

        Gbm_deinit();
        return FALSE;
    }

    if (! get_gbm_numberOfPages(filename, fd, ft, &numPages))
    {
        // only read the first bitmap (fallback)
        numPages = 1;
    }

    /* allocate pages */
    d->pages    = NULL;
    d->numPages = 0;
    try
    {
        d->pages    = new GbmDocumentPage[numPages];
        d->numPages = numPages;
    }
    catch(...)
    {
        d->pages = NULL;
        // error reporting by checking for the NULL pointer is below
    }
    // for compilers that don't throw an std::bad_alloc
    if (d->pages == NULL)
    {
        const char * message       = "Out of memory.";
        const int    messageLength = strlen(message);

        *error = (char *)SOMMalloc(messageLength + 1);
        strcpy(*error, message);

        Gbm_io_close(fd);
        Gbm_deinit();
        return FALSE;
    }

    char indexString[20] = { 0 };

    for (int page = 0; page < numPages; page++)
    {
        GbmDocumentPage & gbmPage(d->pages[page]);
        GBM             * pGbm    = &(gbmPage.gbm);
        GBMRGB          * pGbmRgb =   gbmPage.gbmrgb;

        // default is to load page 0. Old codecs might not support index keyword
        if (page == 0)
        {
            strcpy(indexString, "");
        }
        else
        {
            sprintf(indexString, "index=%d", page);
        }

        rc = load_gbm_page(filename, fd, ft, pGbm, pGbmRgb, indexString, &(gbmPage.data));
        if (rc != GBM_ERR_OK)
        {
            const char * errorMessage       = Gbm_err(rc);
            const int    errorMessageLength = strlen(errorMessage);

            const char * message       = "Can't read bitmap page: %s";
            const int    messageLength = strlen(message);

            *error = (char *)SOMMalloc(messageLength + errorMessageLength + 1);
            sprintf(*error, message, errorMessage);

            Gbm_io_close(fd);
            Gbm_deinit();

            d->deletePages();
            return FALSE;
        }
    }

    Gbm_io_close(fd);
    Gbm_deinit();

    return TRUE;
}

// ---------------------------------------------------------

SOM_Scope short  SOMLINK getBpp(LuGbmDocument *somSelf, Environment *ev)
{
    /* always scale the bitmap up to 24 bits per pixel thus eliminating color palette */
    return 3;
}

// ---------------------------------------------------------

SOM_Scope boolean  SOMLINK isScalable(LuGbmDocument *somSelf,
                                      Environment   *ev)
{
    return TRUE;
}

// ---------------------------------------------------------

SOM_Scope boolean  SOMLINK isRotable(LuGbmDocument *somSelf,
                                     Environment   *ev)
{
    // @todo Add this later if it works in Lucide at all.
    return FALSE;
}

// ---------------------------------------------------------

SOM_Scope long  SOMLINK getPageCount(LuGbmDocument *somSelf,
                                     Environment   *ev)
{
    const LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
    const GbmDocument       *d       = (GbmDocument *)somThis->data;

    return d->numPages;
}

// ---------------------------------------------------------

SOM_Scope void  SOMLINK getPageSize(LuGbmDocument *somSelf,
                                    Environment   *ev     , long     pagenum,
                                    double        *width  , double * height)
{
    const LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
    const GbmDocument       *d       = (GbmDocument *)somThis->data;

    if ((pagenum < 0) || (pagenum >= d->numPages))
    {
        *width  = 0;
        *height = 0;
        return;
    }

    const GbmDocumentPage &page(d->pages[pagenum]);

    if ( width != NULL )
    {
        *width = page.gbm.w;
    }
    if ( height != NULL )
    {
        *height = page.gbm.h;
    }
}

// ---------------------------------------------------------

SOM_Scope void  SOMLINK renderPageToPixbuf(LuGbmDocument *somSelf,
                                           Environment   *ev,
                                           long           pagenum   , long       src_x,
                                           long           src_y     , long       src_width,
                                           long           src_height, double     scale,
                                           long           rotation  , LuPixbuf * pixbuf)
{
    const LuGbmDocumentData *somThis = LuGbmDocumentGetData(somSelf);
    const GbmDocument       *d       = (GbmDocument *)somThis->data;

    if ((scale < 0.0) || (pixbuf == NULL) || (pagenum < 0)    || (pagenum >= d->numPages) ||
        (src_y < 0)   || (src_x < 0)      || (src_width <= 0) || (src_height <= 0))
    {
        return;
    }

    const GbmDocumentPage &page(d->pages[pagenum]);

          int    copySrcWidth  = page.gbm.w;
          int    copySrcHeight = page.gbm.h;
          int    copySrcStride = ((page.gbm.w * page.gbm.bpp + 31)/32) * 4;
    const byte * copySrcData   = page.data;
          byte * scaledData    = NULL;

    // Needs scaling ?
    if (scale != 1.0)
    {
        const int scaledWidth  = (int) (scale * ((double) (page.gbm.w)) + 1.0);
        const int scaledHeight = (int) (scale * ((double) (page.gbm.h)) + 1.0);
        const int scaledStride = ((scaledWidth * page.gbm.bpp + 31)/32) * 4;

        scaledData = (byte *) malloc(scaledStride * scaledHeight);
        if (scaledData != NULL)
        {
            // scale to requested size
            if (gbm_simple_scale(copySrcData, page.gbm.w , page.gbm.h,
                                 scaledData , scaledWidth, scaledHeight,
                                 page.gbm.bpp) == GBM_ERR_OK)
            {
                // set the copy source to the scaled data
                copySrcWidth  = scaledWidth;
                copySrcHeight = scaledHeight;
                copySrcStride = scaledStride;
                copySrcData   = scaledData;
            }
        }
        else
        {
            /* In case we had a problem with scaling (e.g. out of memory), */
            /* render nothing (or later show an error icon?).              */
            DosBeep(500, 100);
            return;
        }
    }

    const byte *src;
          byte *dst;

    int i, y;

    const int   bpp                  = getBpp(somSelf, ev);
    const int   pixbuf_dst_rowstride = pixbuf->getRowSize(ev);
    byte      * pixbuf_dst_data      = (byte *) pixbuf->getDataPtr(ev);

    for ( y = copySrcHeight-(src_y+src_height), i = 0; i < src_height; y++, i++ )
    {
        src = copySrcData     + (y * copySrcStride) + (src_x * bpp);
        dst = pixbuf_dst_data + (i * pixbuf_dst_rowstride);
        memcpy( dst, src, src_width * bpp );
    }

    // cleanup the scale buffer if necessary
    if (scaledData != NULL)
    {
        free(scaledData);
    }
}

// ---------------------------------------------------------
// ---------------------------------------------------------

SOM_Scope void SOMLINK somDefaultInit(LuGbmDocument * somSelf,
                                      som3InitCtrl  * ctrl)
{
    LuGbmDocumentData *somThis;
    somInitCtrl        globalCtrl;
    somBooleanVector   myMask;
    LuGbmDocument_BeginInitializer_somDefaultInit;
    LuGbmDocument_Init_LuDocument_somDefaultInit(somSelf, ctrl);

    // local LuGbmDocument initialization code
    GbmDocument *d = new GbmDocument;
    somThis->data = d;
}

// ---------------------------------------------------------

SOM_Scope void SOMLINK somDestruct(LuGbmDocument *somSelf, octet doFree,
                                   som3DestructCtrl* ctrl)
{
    LuGbmDocumentData *somThis;
    somDestructCtrl    globalCtrl;
    somBooleanVector   myMask;
    LuGbmDocument_BeginDestructor;

    // local LuGbmDocument deinitialization code
    GbmDocument *d = (GbmDocument *)somThis->data;

    delete d;
    d = NULL;

    // end of local LuGbmDocument deinitialization code

    LuGbmDocument_EndDestructor;
}


